# -*- encoding: utf-8 -*-
"""
@File : tools.py
@Time : 2019/03/04 08:35:46
@Author : wangtf
@Version : 1.0
@Desc : None
"""

# here put the import lib
import os
import glob
import numpy as np


def voc_ap(recall, precision, use_07_metric=False):
    """
    ap = voc_ap(recall, precision, [use_07_metric])

    Compute VOC AP given precision and recall.
    If use_07_metric is true, uses  the
    VOC 07 11 point method (default: False).
    Please make shure that recall and precison are sorted by scores.

    Args:
        recall: the shape of (n,) ndarray;
        precision: the shape of (n,) ndarray;
        use_07_metric: if true, the 11 points method will be used.
    Returns:
        the float number result of average precision.
    """
    if use_07_metric:
        # 11 point metric
        ap = 0.
        for t in np.arange(0., 1.1, 0.1):
            if np.sum(recall >= t) == 0:
                p = 0
            else:
                p = np.max(precision[recall >= t])
            ap = ap + p / 11.
    else:
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.], recall, [1.]))
        mpre = np.concatenate(([0.], precision, [0.]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap


def compute_overlaps(boxes, one_box):
    """
    iou = compute_overlaps(boxes, one_box)

    compute intersection over union of ndarray.
    The format of one_box is [xmin, ymin, xmax, ymax].

    Args:
        boxes: the (n, 4) shape ndarray, ground truth boundboxes;
        bb: the (4,) shape ndarray, detected boundboxes;
    Returns:
        a (n, ) shape ndarray.
    """
    # compute overlaps
    # intersection
    ixmin = np.maximum(boxes[:, 0], one_box[0])
    iymin = np.maximum(boxes[:, 1], one_box[1])
    ixmax = np.minimum(boxes[:, 2], one_box[2])
    iymax = np.minimum(boxes[:, 3], one_box[3])
    iw = np.maximum(ixmax - ixmin + 1., 0.)
    ih = np.maximum(iymax - iymin + 1., 0.)
    inters = iw * ih

    # union
    boxes_area = (boxes[:, 2] - boxes[:, 0] + 1.) * (boxes[:, 3] -
                                                     boxes[:, 1] + 1.)
    one_box_area = (one_box[2] - one_box[0] + 1.) * (one_box[3] - one_box[1] +
                                                     1.)
    iou = inters / (one_box_area + boxes_area - inters)

    return iou


def voc_eval(class_recs: dict,
             detect: dict,
             iou_thresh: float = 0.5,
             use_07_metric: bool = False):
    """
    recall, precision, ap = voc_eval(class_recs, detection,
                                [iou_thresh],
                                [use_07_metric])

    Top level function that does the PASCAL VOC evaluation.
    Please make sure that the class_recs only have one class annotations.

    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    Args:
        class_recalls: recalls dict of a class
            class_recs[image_name]={'bbox': []}.
        detection: Path to annotations
            detection={'image_ids':[], bbox': [], 'confidence':[]}.
        [iou_thresh]: Overlap threshold (default = 0.5)
        [use_07_metric]: Whether to use VOC07's 11 point AP computation
            (default False)
    Returns:
        a dict of result including true_positive_number, false_positive_number,
        recall, precision and average_precision.
    Raises:
        TypeError: the data format is not np.ndarray.
    """
    # format data
    # class_rec data load
    npos = 0
    for imagename in class_recs.keys():
        if not isinstance(class_recs[imagename]['bbox'], np.ndarray):
            raise TypeError
        detected_num = class_recs[imagename]['bbox'].shape[0]
        npos += detected_num
        class_recs[imagename]['det'] = [False] * detected_num

    # detections data load
    image_ids = detect['image_ids']
    confidence = detect['confidence']
    BB = detect['bbox']
    if not isinstance(confidence, np.ndarray):
        raise TypeError
    if not isinstance(BB, np.ndarray):
        raise TypeError

    # sort by confidence
    sorted_ind = np.argsort(-confidence)
    BB = BB[sorted_ind, :]
    image_ids = [image_ids[x] for x in sorted_ind]

    # go down dets and mark TPs and FPs
    nd = len(image_ids)
    tp = np.zeros(nd)
    fp = np.zeros(nd)
    for d in range(nd):
        R = class_recs[image_ids[d]]
        bb = BB[d, :].astype(float)
        iou_max = -np.inf
        BBGT = R['bbox'].astype(float)

        if BBGT.size > 0:
            overlaps = compute_overlaps(BBGT, bb)
            iou_max = np.max(overlaps)
            iou_max_index = np.argmax(overlaps)

        if iou_max > iou_thresh:
            if not R['det'][iou_max_index]:
                tp[d] = 1.
                R['det'][iou_max_index] = 1
            else:
                fp[d] = 1.
        else:
            fp[d] = 1.

    # compute precision recall
    fp = np.cumsum(fp)
    tp = np.cumsum(tp)
    true_positive_number = tp[-1]
    false_positive_number = fp[-1]

    recall = tp / float(npos)
    # avoid divide by zero in case the first detection matches a difficult ground truth
    precision = tp / np.maximum(tp + fp, np.finfo(np.float64).eps)
    average_precision = voc_ap(recall, precision, use_07_metric)

    result = {}
    result['true_positive_number'] = true_positive_number
    result['false_positive_number'] = false_positive_number
    result['recall'] = recall
    result['precision'] = precision
    result['average_precision'] = average_precision
    return result


def voc_eval_files(class_recs_dir,
                   detect_file,
                   label_id,
                   iou_thresh=0.5,
                   use_07_metric=False):
    """
    recall, precision, ap = voc_eval(class_recs, detection,
                                [iou_thresh],
                                [use_07_metric])

    Top level function that does the PASCAL VOC evaluation.
    Please make sure that the class_recs only have one class annotations.

    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    Args:
        class_recalls: recalls dict of a class
            class_recs[image_name]={'bbox': []}.
        detection: Path to annotations
            detection={'image_ids':[], bbox': [], 'confidence':[]}.
        [iou_thresh]: Overlap threshold (default = 0.5)
        [use_07_metric]: Whether to use VOC07's 11 point AP computation
            (default False)
    Returns:
        a dict of result including true_positive_number, false_positive_number,
        recall, precision and average_precision.
    Raises:
        IOError: can not find the path.
    """
    if not os.path.exists(class_recs_dir):
        raise IOError
    if not os.path.exists(detect_file):
        raise IOError

    class_recs = {}
    recs_list = glob.glob(os.path.join(class_recs_dir, '*.txt'))
    for path in recs_list:
        image_id = os.path.basename(path)[:-4]
        with open(path) as f:
            data = f.read().strip().split('\n')
            bboxes = []
            for line in data:
                label, xmin, ymin, xmax, ymax = line.strip().split(' ')
                if label == str(label_id):
                    bboxes.append([xmin, ymin, xmax, ymax])
            bboxes = np.array(bboxes)
            class_recs[image_id] = {'bbox': bboxes}

    detection = {'image_ids': [], 'bbox': [], 'confidence': []}
    with open(detect_file) as f:
        data = f.read().strip().split('\n')
        for line in data:
            image_id, confidence, xmin, ymin, xmax, ymax = line.strip().split()
            detection['image_ids'].append(image_id)
            detection['confidence'].append(confidence)
            detection['bbox'].append([xmin, ymin, xmax, ymax])
    detection['image_ids'] = np.array(detection['image_ids'])
    detection['confidence'] = np.array(detection['confidence'])
    detection['bbox'] = np.array(detection['bbox'])

    result = voc_eval(class_recs,
                      detection,
                      iou_thresh=iou_thresh,
                      use_07_metric=use_07_metric)
    return result



#!/usr/bin/env python3

"""
@File: datasplit.py
@Time: 2019-01-11
@Author: ternencewang2015@outlook.com
@Description:
This script used to split pascal voc format dataset and generate
train.txt, val.txt and test.txt.
"""

import os
import glob
import random
import xml.etree.ElementTree as ET


class DataSplit():
    """
    This script used to split pascal voc format dataset and generate
    train.txt, val.txt and test.txt.
    """
    def __init__(self, root_dir):
        """
        Arguments:
        ==========
            root_dir: str, the directory path including JPEGImages
                      and Annotations;
        """
        self.root_dir = root_dir

        self.txt_dir = os.path.join(self.root_dir, 'ImageSets/Main')
        self.xmls_dir = os.path.join(self.root_dir, 'Annotations')
        self.images_dir = os.path.join(self.root_dir, 'JPEGImages')
        assert os.path.exists(self.xmls_dir), self.xmls_dir
        assert os.path.exists(self.images_dir), self.images_dir

        if not os.path.exists(self.txt_dir):
            os.makedirs(self.txt_dir)

        self.useful_name_list = self.match_xml_and_jpg()

    def match_xml_and_jpg(self, xmls_dir=None, images_dir=None):
        """Finding corresponding jpg file for xml file in xmls_dir,
        the list of file name will be returned.
        Arguments:
        ==========
            xmls_dir: str, the directory path including xmls;
            images_dir: str, the directory path incuding images;

        Returns:
        ========
            useful_name_list: str, the list of image name which
                              have corresponding xml file.
        """
        if not xmls_dir:
            xmls_dir = self.xmls_dir
        if not images_dir:
            images_dir = self.images_dir

        xmls_list = sorted(glob.glob(os.path.join(xmls_dir, '*.xml')))
        useful_name_list = []
        for xml_path in xmls_list:
            name = os.path.basename(xml_path).split('.')[0]
            image_path = os.path.join(images_dir, '{}.jpg'.format(name))
            if os.path.exists(image_path):
                useful_name_list.append(name)

        return useful_name_list

    def prefix_grouping(self, prefix_list, name_list=None):
        """
        Arguments:
        ==========
            prefix_list: list, the images which have the same string
                         will save in the same list that as the value
                         of the prefix string as the key.
            name_list: list, default is None, whick have all useful
                         name as the whole dataset.
        Returns:
        ========
            groups: map, the key is the string in prefix_list, the value
                    is a list that all name in it has corresponding prefix.
        """
        if name_list is None:
            name_list = self.useful_name_list

        groups = {}
        for prefix in prefix_list:
            groups[prefix] = []

        for name in name_list:
            for prefix in prefix_list:
                if name.startswith(prefix):
                    groups[prefix].append(name)

        return groups

    def split_by_rate(self, test_rate, val_rate=0.0, name_list=None, shuffle=False):
        """
        Arguments:
        ==========
            test_rate: float, the test data rate for all data;
            val_rate: float, default is 0.0, the val data rate for all data;
            name_list: list, all useful name in this data.
            shuffle: bool, default is False, The name_list will be shuffled
                     if it is true.

        Returns:
        ========
            splited_data: map, the key is str in ['train', 'val', 'test'],
                          the value is the list of names. 
        """
        if name_list is None:
            name_list = self.useful_name_list

        assert test_rate < 1, 'Error: test_rate {} not in range.'.format(test_rate)
        assert len(name_list) > 2, 'Error: name_list length is needed more than 2.'

        if shuffle:
            random.shuffle(name_list)

        test_number = int(test_rate * len(name_list))
        test_number = test_number if test_number > 0 else 1

        if val_rate > 0:
            val_number = int(val_rate * len(name_list))
            val_number = val_number if val_number > 0 else 1
        else:
            val_number = 0

        train_number = len(name_list) - test_number - val_number
        assert train_number > 0, 'Error: train_number is needed more than 0.'
        
        train_list = name_list[0:train_number]
        test_list = name_list[train_number:train_number+test_number]
        if val_number > 0:
            val_list = name_list[-val_number:]
        else:
            val_list = []
        return {'train': train_list, 'val': val_list, 'test': test_list}

    def split_group_by_rate(self, groups, test_rate, val_rate=0.0, shuffle=False):
        """
        Arguments:
        ==========
            groups: dict like {prefix: [name, ]}, grouped name list.
            test_rate: float, the test rate of all data.
            val_rate: float, the val rate of all data, default is 0.0.
            shuffle: bool, default is False.

        Returns:
        ========
            result: dict like {'train': [], 'val': [], 'test': []}.
        """
        train_list, val_list, test_list = [], [], []

        for prefix in groups:
            if not groups[prefix]:
                continue
            split_result = self.split_by_rate(test_rate, val_rate,
                                              name_list=groups[prefix],
                                              shuffle=shuffle)
            train_list = train_list + split_result['train']
            test_list = test_list + split_result['test']
            val_list = val_list + split_result['val']

        return {'train': train_list, 'val': val_list, 'test': test_list}

    def save(self, split_name_dic, save_dir=None):
        """
        Arguments:
        ==========
            split_name_dic: map, the splited result fo dataset.
            save_dir: str, default is None, the path using to save result.
                      if None, the result will saved in Main dir corresponding
                      root dir.
        """
        if save_dir is None:
            save_dir = self.txt_dir

        if 'trainval' not in split_name_dic:
            split_name_dic['trainval'] = split_name_dic['train'] + split_name_dic['val']

        for key, val in split_name_dic.items():
            with open(os.path.join(save_dir, key+'.txt'), 'w') as f:
                f.write('\n'.join(val))

        return 0


class DarknetDataset():
    def __init__(self):
        self.voc_sets = [('2007', 'train'), ('2007', 'val'), ('2007', 'test')]
        self.classes = ["aeroplane", "bicycle", "bird", "boat", "bottle",
                        "bus", "car", "cat", "chair", "cow", "diningtable",
                        "dog", "horse", "motorbike", "person", "pottedplant",
                        "sheep", "sofa", "train", "tvmonitor"]

    def get_classes(self, voc_root_path):
        # The labels.list file mast be in voc_root_path dir.
        label_path = os.path.join(voc_root_path, 'labels.list')
        assert os.path.exists(label_path), label_path

        with open(label_path) as f:
            labels = f.read().strip().split('\n')
            labels = [label.strip() for label in labels]
        self.classes = labels
        print('Classes change to: {}'.format(self.classes))
        return self.classes

    def convert(self, size, box):
        dw = 1./(size[0])
        dh = 1./(size[1])
        x = (box[0] + box[1])/2.0 - 1
        y = (box[2] + box[3])/2.0 - 1
        w = box[1] - box[0]
        h = box[3] - box[2]
        x = x*dw
        w = w*dw
        y = y*dh
        h = h*dh
        return (x, y, w, h)

    def convert_annotation(self, year, image_id, voc_root_path):
        in_file = open(os.path.join(
            voc_root_path, 'VOC%s/Annotations/%s.xml' % (year, image_id)))
        out_file = open(os.path.join(
            voc_root_path, 'VOC%s/labels/%s.txt' % (year, image_id)), 'w')
        tree = ET.parse(in_file)
        root = tree.getroot()
        size = root.find('size')
        w = int(size.find('width').text)
        h = int(size.find('height').text)

        for obj in root.iter('object'):
            difficult = obj.find('difficult').text
            cls = obj.find('name').text
            if cls not in self.classes or int(difficult) == 1:
                continue
            cls_id = self.classes.index(cls)
            xmlbox = obj.find('bndbox')
            b = (float(xmlbox.find('xmin').text), float(xmlbox.find('xmax').text), float(
                xmlbox.find('ymin').text), float(xmlbox.find('ymax').text))
            bb = self.convert((w, h), b)
            out_file.write(str(cls_id) + " " +
                           " ".join([str(a) for a in bb]) + '\n')

    def voc2(self, voc_root_path):
        # change labels
        self.get_classes(voc_root_path)

        for year, image_set in self.voc_sets:
            if not os.path.exists(os.path.join(voc_root_path, 'VOC%s/labels/' % (year))):
                os.makedirs(os.path.join(
                    voc_root_path, 'VOC%s/labels/' % (year)))
            if not os.path.exists(os.path.join(voc_root_path, 'VOC%s/ImageSets/Main/%s.txt' % (year, image_set))):
                continue

            image_ids = open(os.path.join(
                voc_root_path, 'VOC%s/ImageSets/Main/%s.txt' % (year, image_set))).read().strip().split()
            with open(os.path.join(voc_root_path, '%s_%s.txt' % (year, image_set)), 'w') as list_file:
                image_file_list = []
                for image_id in image_ids:
                    image_file_path = os.path.abspath(os.path.join(
                        voc_root_path, 'VOC{}/JPEGImages/{}.jpg'.format(year, image_id)))
                    image_file_list.append(image_file_path)
                    self.convert_annotation(year, image_id, voc_root_path)
                list_file.write('\n'.join(image_file_list))

        #os.system("cat {0}/2007_train.txt {0}/2007_val.txt > {0}/train.txt".format(voc_root_path))
        #os.system("cat {0}/2007_train.txt {0}/2007_val.txt {0}/2007_test.txt > {0}/train.all.txt".format(voc_root_path))


def test():
    root_dir = '/home/wangtf/ShareDataset/dataset/RebarDataset/rebar-test/VOC2007'
    prefix_list = ['12', '14', '16', '18', '20', '22', '25', '32']
    test_rate = 0.2
    val_rate = 0.2

    spliter = DataSplit(root_dir)
    groups = spliter.prefix_grouping(prefix_list)
    split_result = spliter.split_group_by_rate(groups, test_rate, val_rate=val_rate)
    spliter.save(split_result)


if __name__ == '__main__':
    test()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
@File: datasetresize.py
@Time: 2019-01-10 19:05:00
@Author: wangtf
@Version: 1.0
@Dest: resize the whole dataset
"""

import os
import glob
import cv2
import shutil
import tqdm

from ._xml_parser import XmlParser
from .tools import resize_image_by_size


class DatasetResize():
    def __init__(self, root_voc_dir, save_root_dir=None):
        """
        Args:
            root_voc_dir: str, the path of pascal voc dir include VOC2007.
            save_root_dir: str, the path of save path, default path is input path.
        """
        self.root_dir = root_voc_dir
        self.save_root_dir = self.root_dir if save_root_dir is None else save_root_dir

        self.annotations_dir = os.path.join(self.root_dir, 'Annotations')
        assert os.path.exists(self.annotations_dir), self.annotations_dir

        self.images_dir = os.path.join(self.root_dir, 'JPEGImages')
        assert os.path.exists(self.images_dir), self.images_dir

        # set save path
        self.save_annotations_dir = os.path.join(self.save_root_dir, 'Annotations')
        self.save_images_dir = os.path.join(self.save_root_dir, 'JPEGImages')
        if not os.path.exists(self.save_annotations_dir):
            os.makedirs(self.save_annotations_dir)
        if not os.path.exists(self.save_images_dir):
            os.makedirs(self.save_images_dir)

    def get_annotations(self):
        annotations_file_list = glob.glob(os.path.join(self.annotations_dir, '*.xml'))
        return annotations_file_list

    def resize_tuple_by_rate(self, rate, image_path, xml_path, save_image_path=None, save_xml_path=None):
        """Resize a image and coresponding xml
        Arguments:
        ==========
            rate: float or int, scale size.
            image_path: str, the path of image.
            xml_path: str, the path of xml file.
            save_image_path: str, image save path, default is image_path.
            save_xml_path: str, xml save path, default is xml_path.
        """
        assert os.path.exists(image_path), image_path
        assert os.path.exists(xml_path), xml_path

        save_image_path = image_path if save_image_path is None else save_image_path
        save_xml_path = xml_path if save_xml_path is None else save_xml_path

        # resize image and save
        image = cv2.imread(image_path)
        image_resized = cv2.resize(image, None, fx=rate, fy=rate)
        cv2.imwrite(save_image_path, image_resized)

        # resize annotation and save
        parser = XmlParser()
        xml_data = parser.load(xml_path)
        
        original_width = int(xml_data['size']['width'])
        original_height = int(xml_data['size']['height'])

        new_width = int(original_width * rate)
        new_height = int(original_height * rate)

        horizion_bias = 0
        vertical_bias = 0

        xml_data['size']['width'] = new_width
        xml_data['size']['height'] = new_height

        new_objs = []
        for index, obj in enumerate(xml_data['object']):
            xmin = int(float(obj['bndbox']['xmin']))
            ymin = int(float(obj['bndbox']['ymin']))
            xmax = int(float(obj['bndbox']['xmax']))
            ymax = int(float(obj['bndbox']['ymax']))
            
            new_bndbox = self.resize_bbox(rate, horizion_bias, vertical_bias, [xmin, ymin, xmax, ymax])
            new_bndbox = list(map(int, new_bndbox))
            xmin, ymin, xmax, ymax = new_bndbox
            if xmax - xmin < min_obj_size or ymax - ymin < min_obj_size:
                print(f'The new size of {xmin}, {ymin}, {xmax}, {ymax} is smaler than {min_obj_size}*{min_obj_size}. delete')
                continue
            xml_data['object'][index]['bndbox']['xmin'] = str(xmin)
            xml_data['object'][index]['bndbox']['ymin'] = str(ymin)
            xml_data['object'][index]['bndbox']['xmax'] = str(xmax)
            xml_data['object'][index]['bndbox']['ymax'] = str(ymax)
            new_objs.append(xml_data['object'][index])
        xml_data['object'] = new_objs

        save_xml_path = xml_path if save_xml_path is None else save_xml_path
        parser.save(save_xml_path, xml_data)
        return 1

    def resize_tuple_by_min_size(self, min_size, image_path, xml_path, save_image_path=None, save_xml_path=None):
        assert os.path.exists(image_path), image_path
        assert os.path.exists(xml_path), xml_path

        save_image_path = image_path if save_image_path is None else save_image_path
        save_xml_path = xml_path if save_xml_path is None else save_xml_path

        image = cv2.imread(image_path)
        height, width = image.shape[0:2]
        rate = min_size / min(height, width)
        new_height = int(height * rate)
        new_width = int(width * rate)

        self.resize_tuple_by_rate(rate, image_path, xml_path, save_image_path, save_xml_path)

    def resize_tuple_by_size(self, width, height, image_path, xml_path, save_image_path=None, save_xml_path=None, min_obj_size=8):
        assert os.path.exists(image_path), image_path
        assert os.path.exists(xml_path), xml_path

        save_image_path = image_path if save_image_path is None else save_image_path
        save_xml_path = xml_path if save_xml_path is None else save_xml_path

        self.resize_image_by_size(image_path, width, height, save_image_path)
        self.resize_xml_by_size(xml_path, width, height, save_xml_path, min_obj_size=min_obj_size)
        return
    
    def resize_image_by_size(self, image_path, width, height, save_image_path=None):
        assert os.path.exists(image_path), image_path

        image = cv2.imread(image_path)
        resized_image, rate, bias = resize_image_by_size(image, width=width, height=height)

        cv2.imwrite(save_image_path, resized_image)
        assert os.path.exists(save_image_path), 'Result was not saved: '.format(save_image_path)
        return

    def resize_bbox(self, rate, horizion_bias, vertical_bias, bbox):
        xmin, ymin, xmax, ymax = bbox
        xmin = xmin * rate + horizion_bias
        ymin = ymin * rate + vertical_bias
        xmax = xmax * rate + horizion_bias
        ymax = ymax * rate + vertical_bias
        return [xmin, ymin, xmax, ymax]
    
    def resize_xml_by_size(self, xml_path, width, height, save_xml_path=None, min_obj_size=8):
        parser = XmlParser()
        xml_data = parser.load(xml_path)
        
        original_width = int(xml_data['size']['width'])
        original_height = int(xml_data['size']['height'])

        rate = min(float(width) / original_width, float(height) / original_height)
        new_width = int(original_width * rate)
        new_height = int(original_height * rate)

        horizion_bias = int((width - new_width) / 2)
        vertical_bias = int((height - new_height) / 2)

        xml_data['size']['width'] = width
        xml_data['size']['height'] = height

        new_objs = []
        for index, obj in enumerate(xml_data['object']):
            xmin = int(float(obj['bndbox']['xmin']))
            ymin = int(float(obj['bndbox']['ymin']))
            xmax = int(float(obj['bndbox']['xmax']))
            ymax = int(float(obj['bndbox']['ymax']))
            
            new_bndbox = self.resize_bbox(rate, horizion_bias, vertical_bias, [xmin, ymin, xmax, ymax])
            new_bndbox = list(map(int, new_bndbox))
            xmin, ymin, xmax, ymax = new_bndbox
            if xmax - xmin < min_obj_size or ymax - ymin < min_obj_size:
                print(f'The new size of {xmin}, {ymin}, {xmax}, {ymax} is smaler than {min_obj_size}*{min_obj_size}. delete')
                continue
            xml_data['object'][index]['bndbox']['xmin'] = str(xmin)
            xml_data['object'][index]['bndbox']['ymin'] = str(ymin)
            xml_data['object'][index]['bndbox']['xmax'] = str(xmax)
            xml_data['object'][index]['bndbox']['ymax'] = str(ymax)
            new_objs.append(xml_data['object'][index])
        xml_data['object'] = new_objs

        save_xml_path = xml_path if save_xml_path is None else save_xml_path
        parser.save(save_xml_path, xml_data)
        return

    def resize_dataset_by_rate(self, rate):
        """Resize the whole dataset
        Arguments:
        ==========
            rate: float or int, scale size.
        """
        annotations_file_list = self.get_annotations()
        
        print('Resizing dataset ...')
        for xml_path in tqdm.tqdm(annotations_file_list):
            image_path = self.get_image_path_by_xml_path(xml_path)
            save_xml_path = self.get_save_path(xml_path)
            save_image_path = self.get_save_path(image_path)

            self.resize_tuple_by_rate(rate, image_path, xml_path, save_image_path, save_xml_path)

    def resize_dataset_by_min_size(self, min_size):
        annotations_file_list = self.get_annotations()
    
        print('Resizing dataset ...')
        for xml_path in tqdm.tqdm(annotations_file_list):
            image_path = self.get_image_path_by_xml_path(xml_path)
            save_xml_path = self.get_save_path(xml_path)
            save_image_path = self.get_save_path(image_path)

            self.resize_tuple_by_min_size(min_size, image_path, xml_path, save_image_path, save_xml_path)

    def resize_dataset_by_size(self, width, height, min_obj_size=8):
        annotations_file_list = self.get_annotations()
    
        print('Resizing dataset ...')
        for xml_path in tqdm.tqdm(annotations_file_list):
            image_path = self.get_image_path_by_xml_path(xml_path)
            save_xml_path = self.get_save_path(xml_path)
            save_image_path = self.get_save_path(image_path)

            self.resize_tuple_by_size(width, height, image_path, xml_path, save_image_path, save_xml_path, min_obj_size=min_obj_size)
        return

    def get_save_path(self, path):
        name = os.path.basename(path)
        save_dir = self.save_annotations_dir if name[-4:] == '.xml' else self.save_images_dir
        save_path = os.path.join(save_dir, name)
        return save_path
        
    def get_image_path_by_xml_path(self, xml_path):
        xml_name = os.path.basename(xml_path)
        image_path = os.path.join(self.images_dir, xml_name.replace('.xml', '.jpg'))
        return image_path

    def copy_imagesets(self, imagesets_dir=None):
        """Copy some text file in Main dir from root dir to save dir
        Arguments:
        ==========
            images_dir: str, the path of ImageSets/Main.
        """
        if imagesets_dir is None:
            imagesets_dir = os.path.join(self.root_dir, 'ImageSets/Main')

        # make save dir        
        save_dir = os.path.join(self.save_root_dir, 'ImageSets/Main')
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        for file_name in ['train.txt', 'val.txt', 'trainval.txt', 'test.txt']:
            file_path = os.path.join(imagesets_dir, file_name)
            if os.path.exists(file_path):
                shutil.copy2(file_path, save_dir)
            else:
                print('Can not find path: {}'.format(file_path))


def test():
    root_dir = '/diskb/GlodonDataset/SteelPipe/SteelPipe_raw-20190806/VOC-20190929/VOC2007'
    save_root_dir = '/diskb/GlodonDataset/SteelPipe/SteelPipe_raw-20190806/VOC-20190929/VOC2007-test'
    # min_size = 2000
    width = 608
    height = 608
    resizer = DatasetResize(root_dir, save_root_dir)
    # resizer.resize_dataset_by_min_size(min_size)
    resizer.resize_dataset_by_size(width, height)
    resizer.copy_imagesets()


if __name__ == '__main__':
    test()



#!/usr/bin/env python3

import os
import shutil
import glob
import unittest
import sys
sys.path.append('./')
from pascal_voc_tools import format


class TestFormat(unittest.TestCase):
    test_dev_dir = os.path.join(os.path.dirname(__file__), 'devkit')
    save_dir = os.path.join(os.path.dirname(__file__), 'gen_devkit')

    def setUp(self):
        os.makedirs(self.test_dev_dir)
        self.ann_dir = os.path.join(self.test_dev_dir, 'Annotations')
        self.jpeg_dir = os.path.join(self.test_dev_dir, 'JPEGImages')
        os.makedirs(self.ann_dir)
        os.makedirs(self.jpeg_dir)
        
    def tearDown(self):
        shutil.rmtree(self.test_dev_dir)
        shutil.rmtree(self.save_dir)

    def test_gen_voc_format(self):
        voc_dir = os.path.join(self.save_dir, 'VOC2007')
        voc_dir = format.gen_voc_format(self.ann_dir, self.jpeg_dir, voc_dir)

        # dir exists assert
        save_ann_dir = os.path.join(voc_dir, 'Annotations')
        save_jpeg_dir = os.path.join(voc_dir, 'JPEGImages')
        self.assertTrue(os.path.exists(voc_dir))
        self.assertTrue(os.path.exists(save_ann_dir))
        self.assertTrue(os.path.exists(save_jpeg_dir))

        # xml file names assert
        input_xmls = [os.path.basename(path) for path in glob.glob(os.path.join(self.ann_dir, '*.xml'))]
        output_xmls = [os.path.basename(path) for path in glob.glob(os.path.join(save_ann_dir, '*.xml'))]
        self.assertEqual(input_xmls, output_xmls)

        # jpeg file names assert
        input_jpegs = [os.path.basename(path) for path in glob.glob(os.path.join(self.jpeg_dir, '*.jpg'))]
        output_jpegs = [os.path.basename(path) for path in glob.glob(os.path.join(save_jpeg_dir, '*.jpg'))]
        self.assertEqual(input_jpegs, output_jpegs)

    def test_check_devkit_format(self):
        voc_dir = os.path.join(self.save_dir, 'VOC2007')
        voc_dir = format.gen_voc_format(self.ann_dir, self.jpeg_dir, voc_dir)
        result = format.check_devkit_format(self.save_dir)
        self.assertTrue(result)

    def test_check_voc_format(self):
        voc_dir = os.path.join(self.save_dir, 'VOC2007')
        voc_dir = format.gen_voc_format(self.ann_dir, self.jpeg_dir, voc_dir)
        result = format.check_voc_format(voc_dir)
        self.assertTrue(result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os
import unittest
import sys
import glob
from PIL import Image
sys.path.append('./')
from pascal_voc_tools import file

class TestFile(unittest.TestCase):
    voc_dir = os.path.join(__file__, 'VOC2007')
    jpeg_dir = os.path.join(voc_dir, 'JPEGImages')
    xmls_dir = os.path.join(voc_dir, 'Annotations')

    def setUp(self):
        os.makedirs(self.jpeg_dir)
        os.makedirs(self.xmls_dir)

    def test_check_image_format(self):
        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import cv2


def bb_intersection_over_union(boxA, boxB):
    """calculate intersection over union between two boundboxes.

    Arguments:
    ==========
        boxA: list of xmin, ymin, xmax, ymax;
        boxB: list of xmin, ymin, xmax, ymax;
    Return:
    =======
        a float number of iou between two inputs.
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])

    # compute the area of intersection rectangle
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)

    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
    boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = interArea / float(boxAArea + boxBArea - interArea)

    # return the intersection over union value
    return iou


def average_precision_11_point():
    return


def average_precision_matric():
    return


def resize_image_by_size(image, width, height):
    assert len(image.shape) == 3, 'Can only use RGB image.'
    original_height, original_width, channel= image.shape
    mask_image = np.zeros((height, width, channel), dtype=np.uint8)

    rate = min(float(width) / original_width, float(height) / original_height)
    new_width = int(original_width * rate)
    new_height = int(original_height * rate)

    horizion_bias = int((width - new_width) / 2)
    vertical_bias = int((height - new_height) / 2)

    resized_image = cv2.resize(image, (new_width, new_height))
    mask_image[vertical_bias:vertical_bias+new_height, horizion_bias:horizion_bias+new_width] = resized_image

    return mask_image, rate, (horizion_bias, vertical_bias)



"""
Pascal Voc Tools
===============

This package provide some tools using for 
pascal voc format dataset and some usrful
functions.

some function in the below list:
1. XmlParser
2. 
3. DatasetResize
4. DataSplit
5. DarknetDataset
6. SplitImageAnnotation
7. AnnotationTools
8. AnchorsKMeans
9. DarknetConfig
"""

from ._version import version
from .resize import DatasetResize
from .datatools import DataSplit
from .datatools import DarknetDataset
from .image_annotation_split import SplitImageAnnotation
from .annotation_tools import AnnotationTools
from .anchors_kmeans import AnchorsKMeans
from .darknet_config import DarknetConfig
from ._xml_parser import XmlParser

__all__ = ['version', 'DatasetResize',
           'DataSplit', 'DarknetDataset', 'SplitImageAnnotation',
           'AnnotationTools', 'AnchorsKMeans', 'DarknetConfig', 'XmlParser']

name = 'pascal_voc_tools'
__version__ = version




import cv2 
import numpy as np


def draw_rect(im, cords, color = None):
    """Draw the rectangle on the image
    
    Parameters
    ----------
    
    im : numpy.ndarray
        numpy image 
    
    cords: numpy.ndarray
        Numpy array containing bounding boxes of shape `N X 4` where N is the 
        number of bounding boxes and the bounding boxes are represented in the
        format `x1 y1 x2 y2`
        
    Returns
    -------
    
    numpy.ndarray
        numpy image with bounding boxes drawn on it
        
    """
    
    im = im.copy()
    
    cords = cords[:,:4]
    cords = cords.reshape(-1,4)
    if not color:
        color = [255,255,255]
    for cord in cords:
        
        pt1, pt2 = (cord[0], cord[1]) , (cord[2], cord[3])
                
        pt1 = int(pt1[0]), int(pt1[1])
        pt2 = int(pt2[0]), int(pt2[1])
    
        im = cv2.rectangle(im.copy(), pt1, pt2, color, int(max(im.shape[:2])/200))
    return im

def bbox_area(bbox):
    return (bbox[:,2] - bbox[:,0])*(bbox[:,3] - bbox[:,1])
        
def clip_box(bbox, clip_box, alpha):
    """Clip the bounding boxes to the borders of an image
    
    Parameters
    ----------
    
    bbox: numpy.ndarray
        Numpy array containing bounding boxes of shape `N X 4` where N is the 
        number of bounding boxes and the bounding boxes are represented in the
        format `x1 y1 x2 y2`
    
    clip_box: numpy.ndarray
        An array of shape (4,) specifying the diagonal co-ordinates of the image
        The coordinates are represented in the format `x1 y1 x2 y2`
        
    alpha: float
        If the fraction of a bounding box left in the image after being clipped is 
        less than `alpha` the bounding box is dropped. 
    
    Returns
    -------
    
    numpy.ndarray
        Numpy array containing **clipped** bounding boxes of shape `N X 4` where N is the 
        number of bounding boxes left are being clipped and the bounding boxes are represented in the
        format `x1 y1 x2 y2` 
    
    """
    ar_ = (bbox_area(bbox))
    x_min = np.maximum(bbox[:,0], clip_box[0]).reshape(-1,1)
    y_min = np.maximum(bbox[:,1], clip_box[1]).reshape(-1,1)
    x_max = np.minimum(bbox[:,2], clip_box[2]).reshape(-1,1)
    y_max = np.minimum(bbox[:,3], clip_box[3]).reshape(-1,1)
    
    bbox = np.hstack((x_min, y_min, x_max, y_max, bbox[:,4:]))
    
    delta_area = ((ar_ - bbox_area(bbox))/ar_)
    
    mask = (delta_area < (1 - alpha)).astype(int)
    
    bbox = bbox[mask == 1,:]


    return bbox


def rotate_im(image, angle):
    """Rotate the image.
    
    Rotate the image such that the rotated image is enclosed inside the tightest
    rectangle. The area not occupied by the pixels of the original image is colored
    black. 
    
    Parameters
    ----------
    
    image : numpy.ndarray
        numpy image
    
    angle : float
        angle by which the image is to be rotated
    
    Returns
    -------
    
    numpy.ndarray
        Rotated Image
    
    """
    # grab the dimensions of the image and then determine the
    # centre
    (h, w) = image.shape[:2]
    (cX, cY) = (w // 2, h // 2)

    # grab the rotation matrix (applying the negative of the
    # angle to rotate clockwise), then grab the sine and cosine
    # (i.e., the rotation components of the matrix)
    M = cv2.getRotationMatrix2D((cX, cY), angle, 1.0)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])

    # compute the new bounding dimensions of the image
    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))

    # adjust the rotation matrix to take into account translation
    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY

    # perform the actual rotation and return the image
    image = cv2.warpAffine(image, M, (nW, nH))

#    image = cv2.resize(image, (w,h))
    return image

def get_corners(bboxes):
    
    """Get corners of bounding boxes
    
    Parameters
    ----------
    
    bboxes: numpy.ndarray
        Numpy array containing bounding boxes of shape `N X 4` where N is the 
        number of bounding boxes and the bounding boxes are represented in the
        format `x1 y1 x2 y2`
    
    returns
    -------
    
    numpy.ndarray
        Numpy array of shape `N x 8` containing N bounding boxes each described by their 
        corner co-ordinates `x1 y1 x2 y2 x3 y3 x4 y4`      
        
    """
    width = (bboxes[:,2] - bboxes[:,0]).reshape(-1,1)
    height = (bboxes[:,3] - bboxes[:,1]).reshape(-1,1)
    
    x1 = bboxes[:,0].reshape(-1,1)
    y1 = bboxes[:,1].reshape(-1,1)
    
    x2 = x1 + width
    y2 = y1 
    
    x3 = x1
    y3 = y1 + height
    
    x4 = bboxes[:,2].reshape(-1,1)
    y4 = bboxes[:,3].reshape(-1,1)
    
    corners = np.hstack((x1,y1,x2,y2,x3,y3,x4,y4))
    
    return corners

def rotate_box(corners,angle,  cx, cy, h, w):
    
    """Rotate the bounding box.
    
    
    Parameters
    ----------
    
    corners : numpy.ndarray
        Numpy array of shape `N x 8` containing N bounding boxes each described by their 
        corner co-ordinates `x1 y1 x2 y2 x3 y3 x4 y4`
    
    angle : float
        angle by which the image is to be rotated
        
    cx : int
        x coordinate of the center of image (about which the box will be rotated)
        
    cy : int
        y coordinate of the center of image (about which the box will be rotated)
        
    h : int 
        height of the image
        
    w : int 
        width of the image
    
    Returns
    -------
    
    numpy.ndarray
        Numpy array of shape `N x 8` containing N rotated bounding boxes each described by their 
        corner co-ordinates `x1 y1 x2 y2 x3 y3 x4 y4`
    """

    corners = corners.reshape(-1,2)
    corners = np.hstack((corners, np.ones((corners.shape[0],1), dtype = type(corners[0][0]))))
    
    M = cv2.getRotationMatrix2D((cx, cy), angle, 1.0)
    
    
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])
    
    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))
    # adjust the rotation matrix to take into account translation
    M[0, 2] += (nW / 2) - cx
    M[1, 2] += (nH / 2) - cy
    # Prepare the vector to be transformed
    calculated = np.dot(M,corners.T).T
    
    calculated = calculated.reshape(-1,8)
    
    return calculated


def get_enclosing_box(corners):
    """Get an enclosing box for ratated corners of a bounding box
    
    Parameters
    ----------
    
    corners : numpy.ndarray
        Numpy array of shape `N x 8` containing N bounding boxes each described by their 
        corner co-ordinates `x1 y1 x2 y2 x3 y3 x4 y4`  
    
    Returns 
    -------
    
    numpy.ndarray
        Numpy array containing enclosing bounding boxes of shape `N X 4` where N is the 
        number of bounding boxes and the bounding boxes are represented in the
        format `x1 y1 x2 y2`
        
    """
    x_ = corners[:,[0,2,4,6]]
    y_ = corners[:,[1,3,5,7]]
    
    xmin = np.min(x_,1).reshape(-1,1)
    ymin = np.min(y_,1).reshape(-1,1)
    xmax = np.max(x_,1).reshape(-1,1)
    ymax = np.max(y_,1).reshape(-1,1)
    
    final = np.hstack((xmin, ymin, xmax, ymax,corners[:,8:]))
    
    return final


def letterbox_image(img, inp_dim):
    '''resize image with unchanged aspect ratio using padding
    
    Parameters
    ----------
    
    img : numpy.ndarray
        Image 
    
    inp_dim: tuple(int)
        shape of the reszied image
        
    Returns
    -------
    
    numpy.ndarray:
        Resized image
    
    '''

    inp_dim = (inp_dim, inp_dim)
    img_w, img_h = img.shape[1], img.shape[0]
    w, h = inp_dim
    new_w = int(img_w * min(w/img_w, h/img_h))
    new_h = int(img_h * min(w/img_w, h/img_h))
    resized_image = cv2.resize(img, (new_w,new_h))
    
    canvas = np.full((inp_dim[1], inp_dim[0], 3), 0)

    canvas[(h-new_h)//2:(h-new_h)//2 + new_h,(w-new_w)//2:(w-new_w)//2 + new_w,  :] = resized_image
    
    return canvas


#!/usr/bin/env python3

import os
import glob
import shutil
import tqdm


def check_devkit_format(root_dir):
    """
    Check file tree format in Devkit dir.
    Arguments:
    =========
        root_dir:, str, including some VOC**** dir.
    """
    check_result = False
    assert os.path.isdir(root_dir), root_dir
    files = glob.glob(os.path.join(root_dir, '*'))
    for file_path in files:
        if os.path.isfile(file_path):
            continue
        if 'VOC' == os.path.basename(file_path)[:3]:
            check_result = check_voc_format(file_path)
            if not check_result:
                return check_result
    return check_result

def check_voc_format(voc_path):
    """
    Check file tree format in VOC dir like VOC2007.
    Arguments:
    =========
        voc_path: str, the dir path should including Annotations and JPEGImages.
    """
    check_result = True
    assert os.path.exists(voc_path), voc_path
    ann_dir = os.path.join(os.path.join(voc_path, 'Annotations'))
    jpeg_dir = os.path.join(os.path.join(voc_path, 'JPEGImages'))
    
    if not os.path.exists(ann_dir):
        check_result = False
    if not os.path.exists(jpeg_dir):
        check_result = False

    # generate ImageSets/Main dir
    if check_result:
        main_dir = os.path.join(os.path.join(voc_path, 'ImageSets/Main'))
        if not os.path.exists(main_dir):
            os.makedirs(main_dir)
    return check_result

def gen_voc_format(ann_dir, jpeg_dir, save_dir):
    """
    Generate a new format pascal voc data.
    Arguments:
    =========
        ann_dir: str, the path having xmls;
        jpeg_dir: str, the path having images;
        save_dir: str, the path saving file.
    """
    assert os.path.isdir(ann_dir), ann_dir
    assert os.path.isdir(jpeg_dir), jpeg_dir
    while os.path.exists(save_dir):
        print('The dir is existed: {}'.format(save_dir))
        save_dir = input('Please input an new folder:')

    ann_save_dir = os.path.join(save_dir, 'Annotations')
    jpeg_save_dir = os.path.join(save_dir, 'JPEGImages')
    main_save_dir = os.path.join(save_dir, 'ImageSets/Main')

    os.makedirs(ann_save_dir)
    os.makedirs(jpeg_save_dir)
    os.makedirs(main_save_dir)

    xml_file_list = glob.glob(os.path.join(ann_dir, '*.xml'))
    print('Find xml file number: {}'.format(len(xml_file_list)))

    for xml_file in tqdm.tqdm(xml_file_list):
        jpeg_file = os.path.join(jpeg_dir, os.path.basename(xml_file).replace('.xml', '.jpg'))
        assert os.path.isfile(jpeg_file), jpeg_file
        shutil.copy2(jpeg_file, jpeg_save_dir)
        shutil.copy2(xml_file, ann_save_dir)
    print('Done.')
    return save_dir


# Copyright (c) 2017-present, Facebook, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
#
# Based on:
# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------

cimport cython
import numpy as np
cimport numpy as np

cdef inline np.float32_t max(np.float32_t a, np.float32_t b) nogil:
    return a if a >= b else b

cdef inline np.float32_t min(np.float32_t a, np.float32_t b) nogil:
    return a if a <= b else b

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def nms(np.ndarray[np.float32_t, ndim=2] dets, np.float32_t thresh):
    cdef np.ndarray[np.float32_t, ndim=1] x1 = dets[:, 0]
    cdef np.ndarray[np.float32_t, ndim=1] y1 = dets[:, 1]
    cdef np.ndarray[np.float32_t, ndim=1] x2 = dets[:, 2]
    cdef np.ndarray[np.float32_t, ndim=1] y2 = dets[:, 3]
    cdef np.ndarray[np.float32_t, ndim=1] scores = dets[:, 4]

    cdef np.ndarray[np.float32_t, ndim=1] areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    cdef np.ndarray[np.int_t, ndim=1] order = scores.argsort()[::-1]

    cdef int ndets = dets.shape[0]
    cdef np.ndarray[np.int_t, ndim=1] suppressed = \
            np.zeros((ndets), dtype=np.int)

    # nominal indices
    cdef int _i, _j
    # sorted indices
    cdef int i, j
    # temp variables for box i's (the box currently under consideration)
    cdef np.float32_t ix1, iy1, ix2, iy2, iarea
    # variables for computing overlap with box j (lower scoring box)
    cdef np.float32_t xx1, yy1, xx2, yy2
    cdef np.float32_t w, h
    cdef np.float32_t inter, ovr

    with nogil:
      for _i in range(ndets):
          i = order[_i]
          if suppressed[i] == 1:
              continue
          ix1 = x1[i]
          iy1 = y1[i]
          ix2 = x2[i]
          iy2 = y2[i]
          iarea = areas[i]
          for _j in range(_i + 1, ndets):
              j = order[_j]
              if suppressed[j] == 1:
                  continue
              xx1 = max(ix1, x1[j])
              yy1 = max(iy1, y1[j])
              xx2 = min(ix2, x2[j])
              yy2 = min(iy2, y2[j])
              w = max(0.0, xx2 - xx1 + 1)
              h = max(0.0, yy2 - yy1 + 1)
              inter = w * h
              ovr = inter / (iarea + areas[j] - inter)
              if ovr >= thresh:
                  suppressed[j] = 1

    return np.where(suppressed == 0)[0]

# ----------------------------------------------------------
# Soft-NMS: Improving Object Detection With One Line of Code
# Copyright (c) University of Maryland, College Park
# Licensed under The MIT License [see LICENSE for details]
# Written by Navaneeth Bodla and Bharat Singh
# ----------------------------------------------------------
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def soft_nms(
    np.ndarray[float, ndim=2] boxes_in,
    float sigma=0.5,
    float Nt=0.3,
    float threshold=0.001,
    unsigned int method=0
):
    boxes = boxes_in.copy()
    cdef unsigned int N = boxes.shape[0]
    cdef float iw, ih, box_area
    cdef float ua
    cdef int pos = 0
    cdef float maxscore = 0
    cdef int maxpos = 0
    cdef float x1, x2, y1, y2, tx1, tx2, ty1, ty2, ts, area, weight, ov
    inds = np.arange(N)

    for i in range(N):
        maxscore = boxes[i, 4]
        maxpos = i

        tx1 = boxes[i,0]
        ty1 = boxes[i,1]
        tx2 = boxes[i,2]
        ty2 = boxes[i,3]
        ts = boxes[i,4]
        ti = inds[i]

        pos = i + 1
        # get max box
        while pos < N:
            if maxscore < boxes[pos, 4]:
                maxscore = boxes[pos, 4]
                maxpos = pos
            pos = pos + 1

        # add max box as a detection
        boxes[i,0] = boxes[maxpos,0]
        boxes[i,1] = boxes[maxpos,1]
        boxes[i,2] = boxes[maxpos,2]
        boxes[i,3] = boxes[maxpos,3]
        boxes[i,4] = boxes[maxpos,4]
        inds[i] = inds[maxpos]

        # swap ith box with position of max box
        boxes[maxpos,0] = tx1
        boxes[maxpos,1] = ty1
        boxes[maxpos,2] = tx2
        boxes[maxpos,3] = ty2
        boxes[maxpos,4] = ts
        inds[maxpos] = ti

        tx1 = boxes[i,0]
        ty1 = boxes[i,1]
        tx2 = boxes[i,2]
        ty2 = boxes[i,3]
        ts = boxes[i,4]

        pos = i + 1
        # NMS iterations, note that N changes if detection boxes fall below
        # threshold
        while pos < N:
            x1 = boxes[pos, 0]
            y1 = boxes[pos, 1]
            x2 = boxes[pos, 2]
            y2 = boxes[pos, 3]
            s = boxes[pos, 4]

            area = (x2 - x1 + 1) * (y2 - y1 + 1)
            iw = (min(tx2, x2) - max(tx1, x1) + 1)
            if iw > 0:
                ih = (min(ty2, y2) - max(ty1, y1) + 1)
                if ih > 0:
                    ua = float((tx2 - tx1 + 1) * (ty2 - ty1 + 1) + area - iw * ih)
                    ov = iw * ih / ua #iou between max box and detection box

                    if method == 1: # linear
                        if ov > Nt:
                            weight = 1 - ov
                        else:
                            weight = 1
                    elif method == 2: # gaussian
                        weight = np.exp(-(ov * ov)/sigma)
                    else: # original NMS
                        if ov > Nt:
                            weight = 0
                        else:
                            weight = 1

                    boxes[pos, 4] = weight*boxes[pos, 4]

                    # if box score falls below threshold, discard the box by
                    # swapping with last box update N
                    if boxes[pos, 4] < threshold:
                        boxes[pos,0] = boxes[N-1, 0]
                        boxes[pos,1] = boxes[N-1, 1]
                        boxes[pos,2] = boxes[N-1, 2]
                        boxes[pos,3] = boxes[N-1, 3]
                        boxes[pos,4] = boxes[N-1, 4]
                        inds[pos] = inds[N-1]
                        N = N - 1
                        pos = pos - 1

            pos = pos + 1

    return boxes[:N], inds[:N]



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : nms.py
@Time : 2019/03/05 10:35:46
@Author : wangtf
@Version : 1.0
@Desc : None
'''

# here put the import lib
import numpy as np
# import .cython_nms


# def nms(dets, thresh):
#     """Apply classic DPM-style greedy NMS."""
#     if dets.shape[0] == 0:
#         return []
#     return cython_nms.nms(dets, thresh)


# def soft_nms(
#     dets, sigma=0.5, overlap_thresh=0.3, score_thresh=0.001, method='linear'
# ):
#     """Apply the soft NMS algorithm from https://arxiv.org/abs/1704.04503."""
#     if dets.shape[0] == 0:
#         return dets, []

#     methods = {'hard': 0, 'linear': 1, 'gaussian': 2}
#     assert method in methods, 'Unknown soft_nms method: {}'.format(method)

#     dets, keep = cython_nms.soft_nms(
#         np.ascontiguousarray(dets, dtype=np.float32),
#         np.float32(sigma),
#         np.float32(overlap_thresh),
#         np.float32(score_thresh),
#         np.uint8(methods[method])
#     )
#     return dets, keep


def nms(dets, thresh):
    """Apply classic DPM-style greedy NMS."""
    if dets.shape[0] == 0:
        return []

    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    ndets = dets.shape[0]
    suppressed = np.zeros((ndets), dtype=np.int)

    for _i in range(ndets):
        i = order[_i]
        if suppressed[i] == 1:
            continue
        ix1 = x1[i]
        iy1 = y1[i]
        ix2 = x2[i]
        iy2 = y2[i]
        iarea = areas[i]
        for _j in range(_i + 1, ndets):
            j = order[_j]
            if suppressed[j] == 1:
                continue
            xx1 = np.maximum(ix1, x1[j])
            yy1 = np.maximum(iy1, y1[j])
            xx2 = np.minimum(ix2, x2[j])
            yy2 = np.minimum(iy2, y2[j])
            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (iarea + areas[j] - inter)
            if ovr >= thresh:
                suppressed[j] = 1

    return np.where(suppressed == 0)[0]


def soft_nms(boxes_in, sigma=0.5, Nt=0.3, threshold=0.001, method=0):
    """methods = {'hard': 0, 'linear': 1, 'gaussian': 2}
    """
    if boxes_in.shape[0] == 0:
        return boxes_in, []

    boxes = boxes_in.copy()
    N = boxes.shape[0]
    pos = 0
    maxscore = 0
    maxpos = 0
    inds = np.arange(N)

    for i in range(N):
        maxscore = boxes[i, 4]
        maxpos = i

        tx1 = boxes[i, 0]
        ty1 = boxes[i, 1]
        tx2 = boxes[i, 2]
        ty2 = boxes[i, 3]
        ts = boxes[i, 4]
        ti = inds[i]

        pos = i + 1
        # get max box
        while pos < N:
            if maxscore < boxes[pos, 4]:
                maxscore = boxes[pos, 4]
                maxpos = pos
            pos = pos + 1

        # add max box as a detection
        boxes[i, 0] = boxes[maxpos, 0]
        boxes[i, 1] = boxes[maxpos, 1]
        boxes[i, 2] = boxes[maxpos, 2]
        boxes[i, 3] = boxes[maxpos, 3]
        boxes[i, 4] = boxes[maxpos, 4]
        inds[i] = inds[maxpos]

        # swap ith box with position of max box
        boxes[maxpos, 0] = tx1
        boxes[maxpos, 1] = ty1
        boxes[maxpos, 2] = tx2
        boxes[maxpos, 3] = ty2
        boxes[maxpos, 4] = ts
        inds[maxpos] = ti

        tx1 = boxes[i, 0]
        ty1 = boxes[i, 1]
        tx2 = boxes[i, 2]
        ty2 = boxes[i, 3]
        ts = boxes[i, 4]

        pos = i + 1
        # NMS iterations, note that N changes if detection boxes fall below
        # threshold
        while pos < N:
            x1 = boxes[pos, 0]
            y1 = boxes[pos, 1]
            x2 = boxes[pos, 2]
            y2 = boxes[pos, 3]
            s = boxes[pos, 4]

            area = (x2 - x1 + 1) * (y2 - y1 + 1)
            iw = (np.minimum(tx2, x2) - np.maximum(tx1, x1) + 1)
            if iw > 0:
                ih = (np.minimum(ty2, y2) - np.maximum(ty1, y1) + 1)
                if ih > 0:
                    ua = float((tx2 - tx1 + 1) *
                               (ty2 - ty1 + 1) + area - iw * ih)
                    ov = iw * ih / ua  # iou between max box and detection box

                    if method == 1:  # linear
                        if ov > Nt:
                            weight = 1 - ov
                        else:
                            weight = 1
                    elif method == 2:  # gaussian
                        weight = np.exp(-(ov * ov)/sigma)
                    else:  # original NMS
                        if ov > Nt:
                            weight = 0
                        else:
                            weight = 1

                    boxes[pos, 4] = weight*boxes[pos, 4]

                    # if box score falls below threshold, discard the box by
                    # swapping with last box update N
                    if boxes[pos, 4] < threshold:
                        boxes[pos, 0] = boxes[N-1, 0]
                        boxes[pos, 1] = boxes[N-1, 1]
                        boxes[pos, 2] = boxes[N-1, 2]
                        boxes[pos, 3] = boxes[N-1, 3]
                        boxes[pos, 4] = boxes[N-1, 4]
                        inds[pos] = inds[N-1]
                        N = N - 1
                        pos = pos - 1

            pos = pos + 1

    return boxes[:N], inds[:N]




#!/usr/bin/evn python
# -*- coding:utf-8 -*-
"""
@File: _xml_parser.py
@Author: ternencewang
@Desc: decode xml file of Pascal Voc annotation
         write a xml file about pascal voc annotation.
<annotation>
    <folder>{{ folder }}</folder>
    <filename>{{ filename }}</filename>
    <path>{{ path }}</path>
    <source>
        <database>{{ database }}</database>
    </source>
    <size>
        <width>{{ size.width }}</width>
        <height>{{ size.height }}</height>
        <depth>{{ size.depth }}</depth>
    </size>
    <segmented>{{ segmented }}</segmented>{% for obj in object %}
    <object>
        <name>{{ obj.name }}</name>
        <pose>{{ obj.pose }}</pose>
        <truncated>{{ obj.truncated }}</truncated>
        <difficult>{{ obj.difficult }}</difficult>
        <bndbox>
            <xmin>{{ obj.bndbox.xmin }}</xmin>
            <ymin>{{ obj.bndbox.ymin }}</ymin>
            <xmax>{{ obj.bndbox.xmax }}</xmax>
            <ymax>{{ obj.bndbox.ymax }}</ymax>
        </bndbox>
    </object>{% endfor %}
</annotation>

"""
import os
from jinja2 import Environment, PackageLoader
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


class XmlParser():
    """Code and Decode the xml file"""

    def __init__(self):
        self.file_path = None
        self.data = None

        self.template_parameters = {}
        environment = Environment(loader=PackageLoader('pascal_voc_tools',
                                                       'templates'),
                                  keep_trailing_newline=True)
        self.annotation_template = environment.get_template('annotation.xml')
        # some peramaters
        self.size = []

    def load(self, xml_file_path):
        """Load a xml file data.
        Args:
            xml_file_path: str, the path of a xml file
        Returns:
            a dict including info in xml file.
        """
        assert os.path.exists(xml_file_path), xml_file_path

        self.file_path = xml_file_path
        tree = ET.parse(self.file_path)
        root = tree.getroot()
        self.data = self.parse_element(root)
        return self.data

    def replace_name(self, old_name, new_name):
        """Replace an object name.

        Args:
            old_name: str, an object class name.
            new_name: str, a new object class name.
        Raises:
            KeyError: if cannot find 'object' in self.data.
        """
        if 'object' not in self.data:
            raise KeyError('Make shure you have loaded an xml file data.')

        for obj in self.data['object']:
            if obj['name'] == old_name:
                obj['name'] = new_name

    def _save_element_info(self, save_dict, tag, text):
        """Save tag and text to save_dict.
        if tag not in save_dict, it will return like save_dict[tag] = text,
        otherwise it will like save_dict[tag] = [text, ]

        Args:
            save_dict: dict, to save.
            tag: str, the key to save.
            text: str, the value to save.
        """
        if tag not in save_dict:
            if tag != 'object':
                save_dict[tag] = text
            else:
                save_dict[tag] = [text]
        else:
            if not isinstance(save_dict[tag], list):
                save_dict[tag] = [save_dict[tag]]
            save_dict[tag].append(text)

    def parse_element(self, element, save_dict=None):
        """Parse all information in element and save to save_dict.

        Args:
            element: element, an element type node in xml tree.
            save_dict: dict, save result.
        Returns:
            save_dict: dict,
                like {'path': './', 'segmented': '0', 'object': [{}]}.
        """
        if save_dict is None:
            save_dict = {}

        for child in element:
            if not child:
                self._save_element_info(save_dict, child.tag, child.text)
            else:
                self._save_element_info(save_dict, child.tag,
                                        self.parse_element(child))

        return save_dict

    def set_data_head(self,
                      path='',
                      width=0,
                      height=0,
                      depth=3,
                      database='Unknown',
                      segmented=0):
        """Generate a xml file

        Args:
            path: str, arg in xml about image.
            width: int or str, image width.
            height: int or str, image height.
            depth: int or str, image channle.
            database: str, default='Unknown'.
            segmented: int or str, default=0.
        """
        abspath = os.path.abspath(path)

        self.template_parameters['path'] = abspath
        self.template_parameters['filename'] = os.path.basename(abspath)
        self.template_parameters['folder'] = os.path.basename(os.path.dirname(abspath))
        self.template_parameters['size'] = {
                'width': width,
                'height': height,
                'depth': depth,
            }
        self.template_parameters['source'] = {
                'database': database,}
        self.template_parameters['segmented'] = segmented
        self.template_parameters['object'] = []

    def add_object(self,
                   name,
                   xmin,
                   ymin,
                   xmax,
                   ymax,
                   pose='Unspecified',
                   truncated=0,
                   difficult=0):
        """add an object info

        Args:
            name: str, class name.
            xmin: int, left.
            ymin: int, top.
            xmax: int, right.
            ymax: int, bottom.
            pose: str, default is 'Unspecified'.
            truncated: str, default is 0.
            difficult: int, default is 0.
        """
        self.template_parameters['object'].append({
            'name': name,
            'pose': pose,
            'truncated': truncated,
            'difficult': difficult,
            'bndbox': {
                'xmin': xmin,
                'ymin': ymin,
                'xmax': xmax,
                'ymax': ymax,
            },
        })

    def save(self, annotation_path, image_parameters=None):
        """Write a xml file to save info.

        Args:
            annotation_path: str, the path of xml to save.
            image_parameters: dict, some info need to set. Default is None.s
        """
        if image_parameters is not None:
            for k, v in image_parameters.items():
                self.template_parameters[k] = v

        with open(annotation_path, 'w') as xml_file:
            content = self.annotation_template.render(
                **self.template_parameters)
            xml_file.write(content)



from .data_aug import RandomHorizontalFlip
from .data_aug import RandomScale
from .data_aug import RandomTranslate
from .data_aug import RandomRotate
from .data_aug import RandomShear
from .data_aug import Resize
from .data_aug import RandomHSV




#!/usr/bin/env python3

import os
import glob
from PIL import Image
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


def verify_image(jpeg_path):
    """Verify image format.
    Arguments:
    =========
        jpeg_path: str, the image path.
    Return:
    ======
        result: bool, if image ok, True will return.
    """
    assert os.path.exists(jpeg_path), jpeg_path
    try:
        Image.open(jpeg_path).verify()
    except:
        return False
    return True


def check_jpg_xml_match(xml_dir, jpeg_dir):
    """
    Check matching degree about xml files and jpeg files.
    Arguments:
    =========
        xml_dir: str, the dir including xml files;
        jpeg_dir: str, the dir including jpeg files.
    """
    # arguemnts check
    assert os.path.exists(xml_dir), xml_dir
    assert os.path.exists(jpeg_dir), jpeg_dir

    # get name list
    xml_file_list = glob.glob(os.path.join(xml_dir, '*.xml'))
    jpeg_file_list = glob.glob(os.path.join(jpeg_dir, '*.jpg'))
    xml_name_list = [os.path.basename(path).split('.')[0] for path in xml_file_list]
    jpeg_name_list = [os.path.basename(path).split('.')[0] for path in jpeg_file_list]

    inter = list(set(xml_name_list).intersection(set(jpeg_name_list)))
    xml_diff = list(set(xml_name_list).difference(set(jpeg_name_list)))
    jpeg_diff = list(set(jpeg_name_list).difference(set(xml_name_list)))

    # print result and return matched list
    print('Find {} xml, {} jpg, matched {}.'.format(len(xml_file_list), len(jpeg_file_list), len(inter)))
    if len(xml_diff):
        print("Only have xml file: {}\n{}".format(len(xml_diff), xml_diff))
    if len(jpeg_diff):
        print("Only have jpg file: {}\n{}".format(len(jpeg_diff), jpeg_diff))

    return inter


def check_image_format(jpeg_dir):
    """
    Change image format from others to jpg.
    Arguments:
    =========
        jpeg_dir: str, the dir only have image.
    """
    image_path_list = glob.glob(os.path.join(jpeg_dir, '*.*'))
    for jpeg_path in image_path_list:
        suffix = os.path.basename(jpeg_path).split('.')[-1]
        if suffix != 'jpg' and verify_image(jpeg_path):
            image = Image.open(jpeg_path)
            save_path = jpeg_path.replace('.{}'.format(suffix), '.jpg')
            image.save(save_path)
            os.remove(jpeg_path)
            print('Image from {} to {}'.format(jpeg_path, save_path))

    return True


def check_xml_info(xml_info):
    """
    Arguments:
    ==========
        xml_info: dict, including xml data. 
    """
    # check image path
    assert 'path' in xml_info, 'Can not find key(path).'

    image_path = xml_info['path']

    assert os.path.exists(image_path), image_path
    image = Image.open(image_path)
    
    # check image size
    assert 'size' in xml_info, 'Can not find key(size).'
    size = xml_info['size']
    width, height = image.size()
    assert int(size['width']) == width
    assert int(size['height']) == height
    depth = 1 if len(image.mode()) == 1 else 3
    assert int(size['depth']) == depth

    # check object
    objects = xml_info['object']
    for one_obj in objects:
        bndbox = one_obj['bndbox']
        xmin = int(bndbox['xmin'])
        ymin = int(bndbox['ymin'])
        xmax = int(bndbox['xmax'])
        ymax = int(bndbox['ymax'])
        assert xmin > 0, xmin
        assert ymin > 0, ymin
        assert xmax < width, xmax
        assert ymax < height, ymax
